import itertools

import pandas as pd
from openpyxl import load_workbook
from openpyxl.styles import PatternFill


def color_df(df_final, sheet, output_path):
    wb = load_workbook(output_path + "/output/Output file.xlsx")
    ws = wb[sheet]

    # Generate unique region names
    unique_regions = df_final["Regions"].unique()

    # Generate a list of colors (repeat colors if there are more regions than colors)
    colors = ["EA9EBF", "D0A0EC", "B4DFFF", "FFEEBC", "DCEDC1"]
    color_cycle = itertools.cycle(colors)

    # Create a dictionary to map regions to colors
    region_fill_colors = {
        region: PatternFill(start_color=color, end_color=color, fill_type="solid")
        for region, color in zip(unique_regions, color_cycle)
    }

    # Iterate through rows and apply fill colors based on 'Regions' column
    for row in ws.iter_rows(min_row=2, max_row=ws.max_row, min_col=2, max_col=ws.max_column):
        region_value = row[0].value
        region_fill = region_fill_colors.get(region_value, None)

        if region_fill:
            for cell in row:
                cell.fill = region_fill

    wb.save(output_path + "/output/Output file.xlsx")


def save_values(df_final, output_path):
    df_save = df_final.pivot_table(
        index=["Regions", "Mouse ID", "Genotype"],
        columns=["Lipids", "Lipid Class"],
        values=["Values", "Normalized Values"],
        aggfunc="first",
    )
    df_save.reset_index(inplace=True)

    # Save the eliminated lipids and the normalized data with the Z Scores
    with pd.ExcelWriter(output_path + "/output/Output file.xlsx", engine="openpyxl", mode="a") as writer:
        df_save.to_excel(writer, sheet_name="Data for Correlations")

    color_df(df_final, sheet="Data for Correlations", output_path=output_path)


def save_zscores(df_final, output_path):
    df_save = df_final.pivot_table(
        index=["Regions", "Mouse ID", "Genotype"],
        columns=["Lipids", "Lipid Class"],
        values=["Average Z Scores", "Z Scores"],
    )
    df_save.reset_index(inplace=True)

    # Save the eliminated lipids and the normalized data with the Z Scores
    with pd.ExcelWriter(output_path + "/output/Output file.xlsx", engine="openpyxl", mode="a") as writer:
        df_save.to_excel(writer, sheet_name="Z Scores")

    color_df(df_final, sheet="Z Scores", output_path=output_path)

    print("Saving to output file")



import os

import numpy as np
import pandas as pd


def load_data(datapath, sheet_name, mice_sheet):

    print("Loading data from " + datapath)
    # Getting the values for the lipids
    df = pd.read_excel(datapath, sheet_name=sheet_name, header=2)
    df.dropna(axis=1, how="all", inplace=True)

    # Getting the genotypes and regions of the mice samples
    df_mice = pd.read_excel(datapath, sheet_name=mice_sheet, header=None).T
    return df, df_mice


def data_cleanup(df, df_mice, output_path):

    print("Cleaning data")
    print("Removing the Internal Standard samples")
    # Eliminating the 'Internal Standard' samples
    df = df[df["Lipid Class"] != "Internal Standard"]
    # TODO: omega 6 or 3 to remove (DPEA) from all (ask Alain)

    index = df.columns.tolist()
    subjects = []
    lipids = []
    values = []
    regions = []
    genotype = []
    lipid_class = []

    for i, lipid in enumerate(df["Short Name"]):
        for j, subject in enumerate(df.columns[4:]):
            if j != 0:
                y = index.index(j)
                subjects.append(subject)
                lipids.append(lipid)
                lipid_class.append(df.iloc[i, 3])
                values.append(df.iloc[i, y])
                regions.append(df_mice.iloc[1, j])
                genotype.append(df_mice.iloc[0, j])
            else:
                pass

    cleaned_values = [float(value) for value in values]
    lipids = [string.replace("/", "-") for string in lipids]

    df_sorted = pd.DataFrame(
        {
            "Mouse ID": subjects,
            "Lipids": lipids,
            "Lipid Class": lipid_class,
            "Regions": regions,
            "Genotype": genotype,
            "Values": cleaned_values,
        }
    )

    print("Filtering the lipids that have 3 or more values missing")

    # Filter out lipids in the region where they have 3 values missing
    def filter_lipids(df):
        lipid_zero_counts = df.groupby("Lipids")["Values"].apply(lambda x: (x == 0).sum())
        valid_lipids = lipid_zero_counts[lipid_zero_counts < 3].index
        invalid_lipids = lipid_zero_counts[lipid_zero_counts >= 3].index
        valid_df = df[df["Lipids"].isin(valid_lipids)]
        invalid_df = df[df["Lipids"].isin(invalid_lipids)]
        return valid_df, invalid_df

    df_clean = pd.DataFrame()
    df_eliminated = pd.DataFrame()
    for name, group in df_sorted.groupby("Regions"):
        valid_df, invalid_df = filter_lipids(group)
        df_clean = pd.concat([df_clean, valid_df])
        df_eliminated = pd.concat([df_eliminated, invalid_df])

    print("Replacing the zero values with 80% of the minimum value for the corresponding group")

    # Replace zero values with 80% of the minimum value for the corresponding group
    def replace_zero_values(row, data):
        if row["Values"] == 0:
            group_df = data[
                (data["Lipids"] == row["Lipids"])
                & (data["Regions"] == row["Regions"])
                & (data["Genotype"] == row["Genotype"])
                & (data["Values"] != 0)
            ]
            if not group_df.empty:
                min_value = group_df["Values"].min()
                if min_value != 0:
                    new_value = 0.8 * min_value
                    return new_value
        return row["Values"]

    df_clean["Values"] = df_clean.apply(lambda row: replace_zero_values(row, df_clean), axis=1)
    log_values = np.log10(df_clean["Values"])
    df_clean["Normalized Values"] = log_values

    if not os.path.exists(output_path + "/output"):
        os.makedirs(output_path + "/output")

    with pd.ExcelWriter(output_path + "/output/Output file.xlsx") as writer:
        df_eliminated.to_excel(writer, sheet_name="Eliminated Lipids")
    print("Saving data to new Excel file")

    return df_clean



import pandas as pd
import scipy.stats as stats


# TODO: write all of this info into smt


def statistics_tests(df_clean, control_name):

    regions = []
    lipids = []
    shapiro_normality = []
    levene_equality = []

    print("Checking for the normality of the residuals and the equality of the variances")

    # Test for the normality of the residuals and for the equality of variances
    for (region, lipid), data in df_clean.groupby(["Regions", "Lipids"]):
        control_group = data[data["Genotype"] == control_name]
        experimental_group = data[data["Genotype"] != control_name]
        values = data["Normalized Values"]
        shapiro_test = stats.shapiro(values)
        control_data = control_group["Normalized Values"]
        experimental_data = experimental_group["Normalized Values"]
        levene = stats.levene(control_data, experimental_data)

        shapiro_normality.append(shapiro_test.pvalue)
        levene_equality.append(levene.pvalue)
        regions.append(region)
        lipids.append(lipid)

    # Creating a new dataframe with the normality and equality information
    statistics = pd.DataFrame(
        {
            "Regions": regions,
            "Lipids": lipids,
            "Shapiro Normality": shapiro_normality,
            "Levene Equality": levene_equality,
        }
    )

    return statistics


def z_scores(df_clean, statistics):

    print("Computing the Z scores and the average Z scores per lipid class")

    # Z Scores and average Z Scores per lipid class
    grouped = df_clean.groupby(["Regions", "Lipids"])["Normalized Values"].agg(["mean", "std"]).reset_index()
    grouped.rename(columns={"mean": "Mean", "std": "STD"}, inplace=True)
    df_final = pd.merge(df_clean, grouped, on=["Regions", "Lipids"], how="left")
    df_final["Z Scores"] = (df_final["Normalized Values"] - df_final["Mean"]) / df_final["STD"]
    average_z_scores = (
        df_final.groupby(["Regions", "Lipid Class", "Mouse ID"])["Z Scores"].mean().reset_index(name="Average Z Scores")
    )
    df_final = pd.merge(df_final, average_z_scores, on=["Lipid Class", "Regions", "Mouse ID"])
    df_final = pd.merge(df_final, statistics, on=["Regions", "Lipids"], how="left")

    return df_final



"""
Automize Science is a Python package designed to elaborate data into graphs coming from lipid extractions (LC/MS).
Starting from a file containing the **pmol/mg** values per each sample, this package streamlines the process of data
analysis and visualization.
"""

from .graph_constructor import *
from .workflows import data_workflow

__version__ = "1.0.0"



from automize_science.computing_statistics import *
from automize_science.data_cleanup import *
from automize_science.saving import *


def data_workflow(file_path, data_sheet, mice_sheet, output_path, control_name):
    """
    Automatically processes lipidomics data.

    :param file_path: Path of the Excel file containing the data.
    :param data_sheet: Name of the sheet containing the data.
    :param mice_sheet: Name of the sheet containing the information about the subjects.
    :param output_path: Path of where to save the outputs.
    :param control_name: Name of the control subject group.
    """

    df, df_mice = load_data(datapath=file_path, sheet_name=data_sheet, mice_sheet=mice_sheet)
    df_clean = data_cleanup(df=df, df_mice=df_mice, output_path=output_path)
    statistics = statistics_tests(df_clean=df_clean, control_name=control_name)
    df_final = z_scores(df_clean=df_clean, statistics=statistics)
    save_values(df_final=df_final, output_path=output_path)
    save_zscores(df_final=df_final, output_path=output_path)
    wb = load_workbook(output_path + "/output/Output file.xlsx")
    wb.create_sheet(title="Comments")
    wb.save(output_path + "/output/Output file.xlsx")

    return df_final



import os

import matplotlib.pyplot as plt
import scipy.stats as stats
import seaborn as sns
import starbars
from openpyxl import load_workbook

__all__ = [
    "value_graph_lipid_class",
    "value_graph_region",
    "values_graph_lipid",
    "zscore_graph_lipid_class",
    "zscore_graph_region",
    "zscore_graph_lipid",
]


def save_sheet(comment, sheet_name, output_path):

    wb = load_workbook(output_path + "/output/Output file.xlsx")
    if sheet_name not in wb.sheetnames:
        wb.create_sheet(title=sheet_name)
        ws = wb[sheet_name]
        ws.append(comment)
        wb.save(output_path + "/output/Output file.xlsx")
    else:
        ws = wb[sheet_name]
        ws.append(comment)
        wb.save(output_path + "/output/Output file.xlsx")


def get_test(shapiro, levene, control_values, experimental_values):
    if shapiro < 0.05 and levene < 0.05:
        stat, pvalue = stats.ttest_ind(control_values, experimental_values)
        test = "T-Test"
    elif shapiro < 0.05 and levene > 0.05:
        stat, pvalue = stats.ttest_ind(control_values, experimental_values, equal_var=False)
        test = "Welch T-Test"
    elif shapiro > 0.05 and levene > 0.05:
        stat, pvalue = stats.mannwhitneyu(control_values, experimental_values)
        test = "Mann Whitney"
    else:
        pvalue = 0
        test = "no test"

    return pvalue, test


# Graphs by Z scores
def zscore_graph_lipid(df_final, control_name, experimental_name, output_path, palette, show=True):

    if not os.path.exists(output_path + "/output/zscore_graphs/lipid"):
        os.makedirs(output_path + "/output/zscore_graphs/lipid")
    order = [control_name, experimental_name]

    for (region, lipid), data in df_final.groupby(["Regions", "Lipids"]):
        shapiro = data.iloc[0]["Shapiro Normality"]
        levene = data.iloc[0]["Levene Equality"]
        control_group = data[data["Genotype"] == control_name]
        experimental_group = data[data["Genotype"] != control_name]
        control_values = control_group["Z Scores"]
        experimental_values = experimental_group["Z Scores"]

        print(f"Creating graph for {lipid} in {region}")

        sns.boxplot(x="Genotype", y="Z Scores", data=data, order=order, hue="Genotype", palette=palette)
        sns.stripplot(x="Genotype", y="Z Scores", data=data, order=order, color="k", size=4)

        pvalue, test = get_test(shapiro, levene, control_values, experimental_values)
        pairs = [(control_name, experimental_name, pvalue)]
        plt.xlabel("Genotype")
        plt.ylabel("Z Scores")
        plt.title(f"Z Scores Distribution for {lipid} in {region}: {control_name} vs {experimental_name}")
        starbars.draw_annotation(pairs)
        plt.savefig(output_path + "/output/graphs/" + f"Z Scores for {lipid} in {region}.png", dpi=1200)
        if show:
            plt.show()
        plt.close()

        comment = [f"For z scores of {lipid} in {region}, {test} was performed. P-value is {pvalue}."]
        save_sheet(comment, "Comments", output_path)


def zscore_graph_region(df_final, control_name, experimental_name, output_path, palette, show=True):

    if not os.path.exists(output_path + "/output/zscore_graphs/region"):
        os.makedirs(output_path + "/output/zscore_graphs/region")

    for region, data in df_final.groupby("Regions"):
        print(f"Creating graph for: {region}")

        sns.boxplot(x="Lipids", y="Z Scores", hue="Genotype", data=data, palette=palette)
        sns.stripplot(x="Lipids", y="Z Scores", data=data, hue="Genotype", dodge=True, color="k", size=4)

        plt.xlabel("Lipids")
        plt.ylabel("Z Scores")
        plt.title(f"Z Scores Distribution in {region}: {control_name} vs {experimental_name}")
        plt.xticks(rotation=90)

        plt.savefig(output_path + f"/output/zscore_graphs/region/Z-Scores Distribution {region}.png", dpi=1200)
        if show:
            plt.show()
        plt.close()


def zscore_graph_lipid_class(df_final, control_name, experimental_name, output_path, palette, show=True):

    if not os.path.exists(output_path + "/output/zscore_graphs/lipid_class"):
        os.makedirs(output_path + "/output/zscore_graphs/lipid_class")

    for region, data in df_final.groupby("Regions"):
        print(f"Creating lipid classes' graph for {region}")

        sns.boxplot(
            x="Lipid Class",
            y="Z Scores",
            hue="Genotype",
            data=data,
            order=data["Lipid Class"].unique(),
            palette=palette,
        )
        sns.stripplot(x="Lipid Class", y="Z Scores", data=data, dodge=True, color="k", size=4)

        plt.xlabel("Lipid Class")
        plt.ylabel("Z Score")
        plt.title(f"Z Scores Distribution in {region}: {control_name} vs {experimental_name}")
        plt.xticks(rotation=90)

        plt.savefig(output_path + f"/output/zscore_graphs/lipid_class/Z-Scores Distribution {region}.png", dpi=1200)
        if show:
            plt.show()
        plt.close()


# Graphs by values
def values_graph_lipid(df_final, control_name, experimental_name, output_path, palette, show=True):

    if not os.path.exists(output_path + "/output/value_graphs/lipid"):
        os.makedirs(output_path + "/output/value_graphs/lipid")
    order = [control_name, experimental_name]

    for (region, lipid), data in df_final.groupby(["Regions", "Lipids"]):
        shapiro = data.iloc[0]["Shapiro Normality"]
        levene = data.iloc[0]["Levene Equality"]
        control_group = data[data["Genotype"] == control_name]
        experimental_group = data[data["Genotype"] != control_name]
        control_values = control_group["Normalized Values"]
        experimental_values = experimental_group["Normalized Values"]

        print(f"Creating graph for {lipid} in {region}")

        sns.boxplot(x="Genotype", y="Normalized Values", data=data, order=order, hue="Genotype", palette=palette)
        sns.stripplot(x="Genotype", y="Normalized Values", data=data, order=order, color="k", size=4)

        pvalue, test = get_test(shapiro, levene, control_values, experimental_values)
        pairs = [(control_name, experimental_name, pvalue)]
        plt.xlabel("Genotype")
        plt.ylabel("Normalized Values")
        plt.title(f"Normalized Values Distribution for {lipid} in {region}: {control_name} vs {experimental_name}")
        starbars.draw_annotation(pairs)
        plt.savefig(output_path + f"/output/value_graphs/lipid/Normalized Values for {lipid} in {region}.png", dpi=1200)
        if show:
            plt.show()
        plt.close()

        comment = [f"For values of {lipid} in {region}, {test} was performed. P-value is {pvalue}."]
        save_sheet(comment, "Comments", output_path)


def value_graph_region(df_final, control_name, experimental_name, output_path, palette, show=True):

    if not os.path.exists(output_path + "/output/value_graphs/region"):
        os.makedirs(output_path + "/output/value_graphs/region")

    for region, data in df_final.groupby("Regions"):

        print(f"Creating graph for {region}")

        plt.figure(figsize=(16, 13))

        sns.boxplot(x="Lipids", y="Normalized Values", hue="Genotype", data=data, palette=palette)
        sns.stripplot(x="Lipids", y="Normalized Values", data=data, hue="Genotype", dodge=True, color="k", size=4)

        plt.xlabel("Lipids")
        plt.ylabel("Normalized Values")
        plt.title(f"Normalized Values Distribution in {region}: {control_name} vs {experimental_name}")
        plt.xticks(rotation=90)

        plt.savefig(output_path + f"/output/value_graphs/region/Normalized Values Distribution {region}.png", dpi=1200)
        if show:
            plt.show()
        plt.close()


def value_graph_lipid_class(df_final, control_name, experimental_name, output_path, palette, show=True):

    if not os.path.exists(output_path + "/output/value_graphs/lipid_class"):
        os.makedirs(output_path + "/output/value_graphs/lipid_class")

    for region, data in df_final.groupby("Regions"):
        print(f"Creating lipid classes' graph for {region}")

        sns.boxplot(
            x="Lipid Class",
            y="Normalized Values",
            hue="Genotype",
            data=data,
            order=data["Lipid Class"].unique(),
            palette=palette,
        )
        sns.stripplot(x="Lipid Class", y="Normalized Values", data=data, dodge=True, color="k", size=4)

        plt.xlabel("Lipid Class")
        plt.ylabel("Normalized Values")
        plt.title(f"Normalized Values Distribution in {region}: {control_name} vs {experimental_name}")
        plt.xticks(rotation=90)

        plt.savefig(
            output_path + f"/output/value_graphs/lipid_class/Normalized Values Distribution {region}.png", dpi=1200
        )
        if show:
            plt.show()
        plt.close()



from automize_science.graph_constructor import *
from automize_science.workflows import *

file_path = "Dementia project.xlsx"
data_sheet = "Quantification"
mice_sheet = "Sheet1"
output_path = "C:/Users/Elide/Documents/git/automize-science/example"
control_name = "WT"
experimental_name = "FTLD"
palette = "Set2"


df = data_workflow(
    file_path="Dementia project.xlsx",
    data_sheet="Quantification",
    mice_sheet="Sheet1",
    output_path="C:/Users/Elide/Documents/git/automize-science/example",
    control_name="WT",
)

zscore_graph_lipid(
    df_final=df,
    control_name="WT",
    experimental_name="FTLD",
    output_path="C:/Users/Elide/Documents/git/automize-science/example",
    palette="Set2",
    show=True,
)


zscore_graph_lipid_class(
    df_final=df,
    control_name="WT",
    experimental_name="FTLD",
    output_path="C:/Users/Elide/Documents/git/automize-science/example",
    palette="Set2",
    show=True,
)
